@@ -30,7 +30,7 @@ GBLauthors[2] = {
 "\n"
 "Andrew Smith\n"
 "http://littlesvr.ca/misc/contactandrew.php\n"
-"Summer 2005 - Summer 2007\n"
+"2005 - 2018\n"
 "- author and maintainer\n"
 "\n"
 "Barb Czegel\n"
@@ -74,6 +74,7 @@ GBLauthors[2] = {
 "Sebastian Pipping\n"
 "https://blog.hartwork.org/\n"
 "- Patch to fix a crash when opening a file from the \"Recently used\" section of the file browser\n"
+"- Patch to use iniparser 4.1 instead of iniparser 2.17\n"
 "\n"
 "Ernst W. Winter\n"
 "December 2006\n"
